#include <reggie.h>
#include <network_tool.h>
#include <platform.h>
#include <lsockets.h>
#include <stdio.h>
#include <string.h>
#include "tracker.h"
#include "gameinitexit.h"
#include "gametime.h"
#include "carphysics.h"
#include "config.h"
#include "text.h"
#include "gameframe.h"
#include "error.h"

ReggieRef gReggieRef;
int gReggieConnected=false;
int gReggieLastUpdateGInfoVersion;
int gHasCheckedLonelyServer;
float gLastQuery=-1000;
float gConnectionTime;

extern NT_SessionRef gSessionRef;
extern NetworkDatagramRef NT_GetSessionEndpoint(NT_SessionRef ref);

#define kReggieUpdateInterval 1


int CouldMyNATCauseMyServerToBeLonely() {
	NetworkAddress local;
	DatagramGetAddress(NT_GetSessionEndpoint(gSessionRef),&local);
	ReggieNetworkType network = ReggieProbeCheck(gReggieRef,false,&local,nil);
	return((network != kReggieNetworkOffline) && (network <=kReggieNetworkProtective) ? true : false);
}

int DoesReggieThinkJoiningThisGuyWillFail(NetworkAddress *remoteAddress) {
	NetworkAddress local;
	DatagramGetAddress(NT_GetSessionEndpoint(gSessionRef),&local);
	ReggiePrejoinType prejoin = ReggieProbePrejoin(gReggieRef,false,&local,remoteAddress);
//	printf("::::%d\n",prejoin);
	return((prejoin == kReggiePrejoinSorry) ? true : false);
}

int DoesReggieNeedToHelpMeJoinThisGuy(NetworkAddress *remoteAddress) {
	NetworkAddress local;
	DatagramGetAddress(NT_GetSessionEndpoint(gSessionRef),&local);
	ReggiePrejoinType prejoin = ReggieProbePrejoin(gReggieRef,false,&local,remoteAddress);
	return((prejoin != kReggiePrejoinNormal) ? true : false);
}

#define kMessageInputBufferSize 8
#define kMaxMessageSize 1024
char gMessageInputBuffer[kMessageInputBufferSize][kMaxMessageSize];
int gMessageInputBufferPos=0;

enum{
	kReggieAdvGameChannel=1,
	kReggieSearchGameChannel,
	kReggieSearchPlayersChannel
};

void TrackerMessageSend(char *message)
{
	if(gReggieConnected)
	{
		HashRef hashRef = HashCreate(kHashFlagReggieKeys);
		HashAppend(hashRef, "$KIND", "REDLINE");
		HashAppend(hashRef, "$MESSAGE", message);
		
		ReggieMessageSend(gReggieRef,nil,hashRef);
		HashDispose(hashRef); 
		if(gMessageInputBufferPos<kMessageInputBufferSize)
			strcpy(gMessageInputBuffer[gMessageInputBufferPos++],message);
	}
}

int TrackerMessageReceive(char *message)
{
	if(gReggieConnected)
	{
		HashRef hash=ReggieMessageCheck(gReggieRef);
		if(hash)
		{
			char *hashResult=nil;
			do{
				hashResult=HashLookup(hash,"$MESSAGE",hashResult);
				if(hashResult)
					if(gMessageInputBufferPos<kMessageInputBufferSize&&strlen(hashResult)<kMaxMessageSize)
					{
						int dupe=false;
						for(int i=0;i<gMessageInputBufferPos;i++)
							if(strcmp(gMessageInputBuffer[gMessageInputBufferPos++],hashResult)==0)
								dupe=true;
						if(!dupe)
							strcpy(gMessageInputBuffer[gMessageInputBufferPos++],hashResult);
					}
			}while(hashResult);
		}
		if(gMessageInputBufferPos)
		{
			strcpy(message,gMessageInputBuffer[--gMessageInputBufferPos]);
			return true;
		}
	}
	return false;
}

void NetworkStartAdvertising()
{
	if(gConfig->trackerEnable)
	{
		if(!ReggieCreateShared(&gReggieRef,NULL,NULL,NT_GetSessionEndpoint(gSessionRef),NULL,"REDLINE",gConfig->playerName,""))
		{
			gReggieConnected=true;
			gReggieLastUpdateGInfoVersion=-1;
			gHasCheckedLonelyServer=false;
			gConnectionTime=0;
		}
	}
}

void NetworkAdvertiseIdle(tGameInfo *gInfo,int *showLonelyMessage,int locked,int forceUpdate)
{
	if(showLonelyMessage)
		*showLonelyMessage=false;
	if(gReggieConnected)
	{
		UInt32 state=ReggieGetStatus(gReggieRef,NULL,NULL);
		if(state==kReggieStateOnline)
		{
			if(gConnectionTime==0)
				gConnectionTime=TimeGetSeconds();
			if(gReggieLastUpdateGInfoVersion!=gInfo->version||forceUpdate)
			{
				char hashString[1024];				
				tMapInfo *mapInfo=(tMapInfo*)FileGetParsedDataPtr(gInfo->map,kParserTypeMapInfoDesc,sizeof(tMapInfo));
				while(char *ch=strchr(gConfig->gameName,'\''))
					*ch='`';
				sprintf(hashString,"$REDLINE$HOSTING=1 $REDLINE$PASSWORD='%d' $REDLINE$NUMPLAYERS=%d $REDLINE$NUMNETPLAYERS=%d $REDLINE$MAXPLAYERS=%d $REDLINE$ARCADE=%d $REDLINE$LOCKED=%d",strlen(gConfig->password)?1:0,gInfo->numPlayers,gInfo->numNetPlayers,gConfig->maxPlayers,gInfo->arcade,locked);
				HashRef hashRef = HashInitialize(kHashFlagReggieKeys,hashString);
				HashAppend(hashRef, "$MESSAGE", gConfig->gameName);
				HashAppend(hashRef, "$REDLINE$MAP", mapInfo->name);

				for(int i=0;i<gInfo->numPlayers;i++)
				{
					NT_MemberInfo memberInfo;
					char key[256];
					sprintf(key,"$REDLINE$PLAYER%dNAME",i);
					HashAppend(hashRef, key, gInfo->playerNames[i]);
					if(gInfo->playerCars[i]!=-1)
					{
						tCarDefinition *car=(tCarDefinition*)FileGetParsedDataPtr(gInfo->playerCars[i],kParserTypeCarDesc,sizeof(tCarDefinition));
						sprintf(key,"$REDLINE$PLAYER%dCAR",i);
						HashAppend(hashRef, key, car->carName);
					}
					if(i>0&&i<gInfo->numNetPlayers)
					{
						if(NT_GetMemberInfo(gSessionRef,gInfo->netID[i],&memberInfo))
						{
							char address[256];
							AddressTranscribe(&memberInfo.address,address);
							sprintf(key,"$ADDRESS$PLAYER%d",i);						
							HashAppend(hashRef, key, address);
						}
					}
					else if(i==0)
					{
						NetworkAddress local,publicAdd;
						char address[256];
						DatagramGetAddress(NT_GetSessionEndpoint(gSessionRef),&local);
						ReggieProbeCheck(gReggieRef,false,&local,&publicAdd);
						if(!AddressIsEmpty(&publicAdd))
						{
							AddressTranscribe(&publicAdd,address);	
							HashAppend(hashRef, "$ADDRESS$PLAYER0", address);
						}
					}
				}
				
				ReggieInsertBegin(gReggieRef,kReggieAdvGameChannel, kReggieRecordAdvertise ,hashRef);/* Makes own copy of hash */
				HashDispose(hashRef); /* Not needed any more */
				gReggieLastUpdateGInfoVersion=gInfo->version;
			}
			if(!gHasCheckedLonelyServer)
				if(TimeGetSeconds()>gConnectionTime+3)
					if(showLonelyMessage)
					{
						*showLonelyMessage=CouldMyNATCauseMyServerToBeLonely();
						printf("CouldMyNATCauseMyServerToBeLonely sez %d\n",*showLonelyMessage);
						gHasCheckedLonelyServer=true;
					}
		}
		else if(state!=kReggieStatePending)
			NetworkStopAdvertising();
	}
	else
		NetworkStartAdvertising();
}

void NetworkStopAdvertising()
{
	if(gReggieConnected)
	{
		ReggieChannelClose(gReggieRef,kReggieAdvGameChannel);
		ReggieDispose(gReggieRef,true);
		gReggieConnected=false;
	}
}


void QueryGameSearch()
{
	HashRef hashRef = HashInitialize(kHashFlagReggieKeys, "$KIND=redline $REDLINE$HOSTING=1");
	ReggieSearchBegin(gReggieRef,kReggieSearchGameChannel,kReggieRecordAdvertise,kReggieSearchFlagNotify,hashRef); /* Makes own copy of hash */
	HashDispose(hashRef); /* Not needed any more */
}

void QueryPlayerSearch()
{
	HashRef hashRef = HashInitialize(kHashFlagReggieKeys,"$KIND=redline $REDLINE$HOSTING=0");
	ReggieSearchBegin(gReggieRef,kReggieSearchPlayersChannel,kReggieRecordAdvertise,kReggieSearchFlagNotify,hashRef); 
	HashDispose(hashRef); /* Not needed any more */
}

void AdvertisePlayer()
{
	char hashString[1024];				
	sprintf(hashString,"$REDLINE$HOSTING=0");
	HashRef hashRef = HashInitialize(kHashFlagReggieKeys,hashString);
	HashAppend(hashRef, "$NAME", gConfig->playerName);
	ReggieInsertBegin(gReggieRef,kReggieAdvGameChannel,kReggieRecordAdvertise,hashRef);
	HashDispose(hashRef);
}

void NetworkSearchGamesInit()
{
	gLastQuery=0;
	if(!ReggieCreateShared(&gReggieRef,NULL,NULL,NT_GetSessionEndpoint(gSessionRef),NULL,"redline",gConfig->playerName,""))
	{
		QueryGameSearch();
		QueryPlayerSearch();
		AdvertisePlayer();
		gReggieConnected=true;
	}
}

void NetworkSearchGames(int *updated,tGameListEntry *games,int *numGames,int maxGames,tGameListPlayerEntry *players,int *numPlayers,int maxPlayers)
{
	static UInt32 lastGameStamp=-1,lastPlayerStamp=-1;
	static float gLastNATCheck=-1;
	*updated=false;
	
	if(gReggieConnected)
	{
		UInt32 state=ReggieGetStatus(gReggieRef,NULL,NULL);
		if(!(state==kReggieStateRefused||state==kReggieStateOffline))
		{
	        SInt32 reggieCount,totalCount;
			UInt32 reggieStamp;
	        Bool8 outdated;
	        reggieStamp=ReggieSearchStatus(gReggieRef,kReggieSearchGameChannel,&reggieCount,&totalCount,&outdated);
			if(reggieStamp!=lastGameStamp)
			{
				if(reggieCount>-1)
				{
					int oldNumGames=*numGames;
					*numGames=0;
					lastGameStamp=reggieStamp;
					*updated=true;
					
					int index=0;
					for(;index<reggieCount;index++) 
					{
						HashRef hashRef = ReggieSearchIndex(gReggieRef, kReggieSearchGameChannel ,index);
						if(hashRef&&*numGames<maxGames) 
						if(HashLookup(hashRef, "$UNIQUEID", nil))
						{
							Char8 *hashLookupResult;
							
							strcpy(games[*numGames].host,"");
							strcpy(games[*numGames].map,"");
							strcpy(games[*numGames].name,"");
							games[*numGames].numPlayers=0;
							games[*numGames].locked=false;
							games[*numGames].numNetPlayers=0;
							games[*numGames].arcade=false;
							games[*numGames].password=false;
							games[*numGames].loaded=true;

							if(hashLookupResult=HashLookup(hashRef, "$ADDRESS", nil))
								strcpy(games[*numGames].host,hashLookupResult);
																	
							if(hashLookupResult=HashLookup(hashRef, "$REDLINE$MAP", nil))
								strcpy(games[*numGames].map,hashLookupResult);
								
							if(hashLookupResult=HashLookup(hashRef, "$REDLINE$NUMPLAYERS", nil))
								sscanf(hashLookupResult,"%d",&games[*numGames].numPlayers);
								
							if(hashLookupResult=HashLookup(hashRef, "$REDLINE$NUMNETPLAYERS", nil))
								sscanf(hashLookupResult,"%d",&games[*numGames].numNetPlayers);

							if(hashLookupResult=HashLookup(hashRef, "$REDLINE$MAXPLAYERS", nil))
								sscanf(hashLookupResult,"%d",&games[*numGames].maxPlayers);

							if(hashLookupResult=HashLookup(hashRef, "$REDLINE$ARCADE", nil))
								sscanf(hashLookupResult,"%d",&games[*numGames].arcade);
								
							if(hashLookupResult=HashLookup(hashRef, "$REDLINE$PASSWORD", nil))
								sscanf(hashLookupResult,"%d",&games[*numGames].password);
							
							if(hashLookupResult=HashLookup(hashRef, "$REDLINE$LOCKED", nil))
								sscanf(hashLookupResult,"%d",&games[*numGames].locked);

							if(hashLookupResult=HashLookup(hashRef, "$MESSAGE", nil))	
								strcpy(games[*numGames].name,hashLookupResult);
							
							for(int i=0;i<games[*numGames].numPlayers;i++)
							{
								strcpy(games[*numGames].players[i].name,"");
								strcpy(games[*numGames].players[i].car,"");
								strcpy(games[*numGames].players[i].location,"");

								char hashString[32];
								sprintf(hashString,"$REDLINE$PLAYER%dNAME",i);
								if(hashLookupResult=HashLookup(hashRef,hashString,nil))
									strcpy(games[*numGames].players[i].name,hashLookupResult);
									
								sprintf(hashString,"$REDLINE$PLAYER%dCAR",i);
								if(hashLookupResult=HashLookup(hashRef,hashString,nil))
									strcpy(games[*numGames].players[i].car,hashLookupResult);	      					      			

								sprintf(hashString,"$LOCATION$PLAYER%d",i);
								if(hashLookupResult=HashLookup(hashRef,hashString,nil))
									strcpy(games[*numGames].players[i].location,hashLookupResult);
								else
								{
									sprintf(hashString,"$DOMAIN$PLAYER%d",i);
									if(hashLookupResult=HashLookup(hashRef,hashString,nil))
										strcpy(games[*numGames].players[i].location,hashLookupResult);
									else
									{
										sprintf(hashString,"$HOSTNAMEPLAYER%d",i);
										if(hashLookupResult=HashLookup(hashRef,hashString,nil))
											strcpy(games[*numGames].players[i].location,hashLookupResult);
									}
								}
							}

							

							(*numGames)++;
						}
					}
					for(;index<totalCount;index++)
						if(*numGames<maxGames)
						{
							if(oldNumGames<*numGames)		        	
								games[*numGames].loaded=false;						
							(*numGames)++;
						}
		        }
		        if(outdated)
		        	QueryGameSearch();
	        }
			else if(TimeGetSeconds()>gLastNATCheck+1)
			{
				gLastNATCheck=TimeGetSeconds();
				for(int i=0;i<*numGames;i++)
				if(games[i].loaded)
				{
					NetworkAddress addy;
					Result32 err=SimpleResolve(eNetworkStackTCPIP,games[i].host,&addy);
					if(DoesReggieNeedToHelpMeJoinThisGuy(&addy))
						games[i].joinState=kJoinStateReggie;
					if(DoesReggieThinkJoiningThisGuyWillFail(&addy))
						games[i].joinState=kJoinStateFail;
				}
			}
			
			reggieStamp=ReggieSearchStatus(gReggieRef,kReggieSearchPlayersChannel,&reggieCount,&totalCount,&outdated);
			if(reggieStamp!=lastPlayerStamp)
			{
				if(reggieCount>-1)
				{
					int oldNumPlayers=*numPlayers;
					*numPlayers=0;
					lastPlayerStamp=reggieStamp;
					*updated=true;
					int index=0;
					for(;index<reggieCount;index++) 
					{
						HashRef hashRef = ReggieSearchIndex(gReggieRef, kReggieSearchPlayersChannel ,index);
						if(hashRef&&*numPlayers<maxPlayers) 
						if(HashLookup(hashRef, "$UNIQUEID", nil))
						{
							Char8 *hashLookupResult;
							
							strcpy(players[*numPlayers].car,"");
							strcpy(players[*numPlayers].location,"");
							strcpy(players[*numPlayers].name,"");

							if(hashLookupResult=HashLookup(hashRef, "$NAME", nil))
								strcpy(players[*numPlayers].name,hashLookupResult);

							if(hashLookupResult=HashLookup(hashRef, "$LOCATION", nil))
								strcpy(players[*numPlayers].location,hashLookupResult);
							
							(*numPlayers)++;
						}
					}
					for(;index<totalCount;index++)
						if(*numPlayers<maxPlayers)
						{
							if(oldNumPlayers<*numPlayers)		        	
							{
								strcpy(players[*numPlayers].car,"");
								strcpy(players[*numPlayers].location,"");
								strcpy(players[*numPlayers].name,"");						
							}
							(*numPlayers)++;
						}
					//printf("%d,%d=>%d\n",reggieCount,totalCount,*numPlayers);
		        }
		        if(outdated)
		        	QueryPlayerSearch();
	        }
		}
		else
			NetworkSearchGamesDone();
	}
	else
	{
		*numGames=0;
		NetworkSearchGamesInit();
	}
}

void NetworkSearchGamesDone()
{
	if(gReggieConnected)
	{
		ReggieChannelClose(gReggieRef,kReggieAdvGameChannel);
		ReggieChannelClose(gReggieRef, kReggieSearchGameChannel);
		ReggieChannelClose(gReggieRef, kReggieSearchPlayersChannel);
		ReggieDispose(gReggieRef,true);
		gReggieConnected=false;
	}
}

void NetworkSearchGamesStop()
{
	if(gReggieConnected)
	{
		ReggieChannelClose(gReggieRef,kReggieAdvGameChannel);
		ReggieChannelClose(gReggieRef, kReggieSearchGameChannel);
		ReggieChannelClose(gReggieRef, kReggieSearchPlayersChannel);
	}
}

int gWaitingForRegistry=false;
int gWaitingForRecord=false;

void TrackerFetchRecord(tFileRef track,tFileRef car,int arcade)
{
	if(gConfig->registerLapTimes)
		if(!ReggieCreate(&gReggieRef,NULL,NULL,NULL,false,"REDLINE",gConfig->playerName,""))
		{
			char hashString[512];
			HashRef hashRef;				
			tMapInfo *mapInfo=(tMapInfo*)FileGetParsedDataPtr(track,kParserTypeMapInfoDesc,sizeof(tMapInfo));
			if(!mapInfo->builtIn)return;
			tCarDefinition *carDef=(tCarDefinition*)FileGetParsedDataPtr(car,kParserTypeCarDesc,sizeof(tCarDefinition));
			if(!carDef->builtIn)return;
					
			sprintf(hashString,"$REDLINE$MAP='%s' $REDLINE$MODE='%d' SORT='+$REDLINE$TIME' LIMIT=1",mapInfo->name,arcade);		

		  	hashRef = HashInitialize(kHashFlagReggieKeys, hashString);
			HashAppend(hashRef, "$REDLINE$CAR", carDef->carName);

			//printf("hash:%s\n",(char*)(*HashSave(hashRef)));
		  	ReggieSearchBegin(gReggieRef,3,kReggieRecordHighScore,0,hashRef); 
		    HashDispose(hashRef);			
			gWaitingForRecord=true;

		}
		else
			TextPrintfToBufferFormatedFading(Vector(0,-0.2),0.1,kTextAlignMiddle,1,2,"Unable to Connect");		
}

void TrackerRegisterLapTime(tFileRef track,tFileRef car,int time,int arcade)
{
	if(gConfig->registerLapTimes)
		if(!ReggieCreate(&gReggieRef,NULL,NULL,NULL,false,"REDLINE",gConfig->playerName,""))
		{
			char hashString[512];
			HashRef hashRef;				
			tMapInfo *mapInfo=(tMapInfo*)FileGetParsedDataPtr(track,kParserTypeMapInfoDesc,sizeof(tMapInfo));
			if(!mapInfo->builtIn)return;
			tCarDefinition *carDef=(tCarDefinition*)FileGetParsedDataPtr(car,kParserTypeCarDesc,sizeof(tCarDefinition));
			if(!carDef->builtIn)return;
					
			sprintf(hashString,"$REDLINE$TIME='%d' $REDLINE$MODE='%d'",time,arcade);		
			hashRef = HashInitialize(kHashFlagReggieKeys,hashString);
			HashAppend(hashRef, "$REDLINE$MAP", mapInfo->name);
			HashAppend(hashRef, "$REDLINE$CAR", carDef->carName);
			HashAppend(hashRef, "$REDLINE$NAME", gConfig->playerName);

			ReggieInsertBegin(gReggieRef, 1, kReggieRecordHighScore ,hashRef);
			HashDispose(hashRef);

			sprintf(hashString,"$REDLINE$TIME<='%d' $REDLINE$MODE='%d' COUNT",time,arcade);		
		  	hashRef = HashInitialize(kHashFlagReggieKeys, hashString);
			HashAppend(hashRef, "$REDLINE$MAP", mapInfo->name);
			HashAppend(hashRef, "$REDLINE$CAR", carDef->carName);
		  	ReggieSearchBegin(gReggieRef,2,kReggieRecordHighScore,kReggieSearchFlagCount,hashRef); 
		    HashDispose(hashRef);
			gWaitingForRegistry=true;

			sprintf(hashString,"$REDLINE$MODE='%d' SORT='+$REDLINE$TIME' LIMIT=1",arcade);		
		  	hashRef = HashInitialize(kHashFlagReggieKeys, hashString);
			HashAppend(hashRef, "$REDLINE$MAP", mapInfo->name);
			HashAppend(hashRef, "$REDLINE$CAR", carDef->carName);
		  	ReggieSearchBegin(gReggieRef,3,kReggieRecordHighScore,0,hashRef); 
		    HashDispose(hashRef);			
			gWaitingForRecord=true;

		}
		else
			TextPrintfToBufferFormatedFading(Vector(0,-0.2),0.1,kTextAlignMiddle,1,2,"Unable to Connect");		
}

void TrackerLapTimeRegistryClose()
{
	if(gWaitingForRegistry||gWaitingForRecord)
	{
		ReggieChannelClose(gReggieRef,1);
		ReggieChannelClose(gReggieRef,2);
		ReggieChannelClose(gReggieRef,3);
		ReggieDispose(gReggieRef,true);
		gWaitingForRegistry=gWaitingForRecord=false;
	}
}

void TrackerWaitForLapTimeRegistry()
{
	if(gWaitingForRegistry||gWaitingForRecord)
	{
		if(gWaitingForRegistry)
			TextPrintfToBufferFormated(Vector(0,-0.2),0.05,kTextAlignMiddle,"Registering Lap Time");

		UInt32 state=ReggieGetStatus(gReggieRef,NULL,NULL);
		if(state==kReggieStateRefused||state==kReggieStateOffline)
		{
			if(state==kReggieStateRefused)
				TextPrintfToBufferFormatedFading(Vector(0,-0.2),0.1,kTextAlignMiddle,1,2,"Connection Refused");
			else
				TextPrintfToBufferFormatedFading(Vector(0,-0.2),0.1,kTextAlignMiddle,1,2,"No Response");
			gWaitingForRegistry=false;
			gWaitingForRecord=false;
		}
		else
		{
			if(gWaitingForRegistry)
			{
		        SInt32 reggieCount,totalCount;
		        ReggieSearchStatus(gReggieRef,2,&reggieCount,&totalCount,NULL);		
		        if(reggieCount!=-1)
				{
					if(totalCount==0)
						TextPrintfToBufferFormatedFading(Vector(0,-0.2),0.1,kTextAlignMiddle,1,2,"World Record!!!");
					else if(totalCount<=25)	
						TextPrintfToBufferFormatedFading(Vector(0,-0.2),0.1,kTextAlignMiddle,1,2,"Place %d on World List",totalCount+1);
					gWaitingForRegistry=false;
				}    
			} 
			if(gWaitingForRecord)
			{
		        SInt32 reggieCount,totalCount;
		        ReggieSearchStatus(gReggieRef,3,&reggieCount,&totalCount,NULL);		
		        if(reggieCount!=-1)
				{
					if(totalCount==0)
					{
						gWorldRecord=0;
						gWaitingForRecord=false;
					}
					else if(reggieCount>0){
						HashRef hashRef = ReggieSearchIndex(gReggieRef,3,0);
						//printf("$REDLINE$TIME%s\n",HashLookup(hashRef,"$REDLINE$TIME",nil));
						//printf("$REDLINE$NAME%s\n",HashLookup(hashRef,"$REDLINE$NAME",nil));
						//printf("hash:%s\n",(char*)(*HashSave(hashRef)));
		      			sscanf(HashLookup(hashRef,"$REDLINE$TIME",nil),"%d",&gWorldRecord);
		      			sscanf(HashLookup(hashRef,"$REDLINE$NAME",nil),"%s",&gRecordName);
						gWaitingForRecord=false;
					}	
				}    
			}
		}
		
		if(!(gWaitingForRegistry||gWaitingForRecord))
		{
			ReggieChannelClose(gReggieRef,1);
			ReggieChannelClose(gReggieRef,2);
			ReggieChannelClose(gReggieRef,3);
			ReggieDispose(gReggieRef,true);
		}   	
	}
}